# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2009- Spyder Project Contributors
#
# Distributed under the terms of the MIT License
# (see spyder/__init__.py for details)
# -----------------------------------------------------------------------------


"""
spyder.plugins.console
======================

Internal Console Plugin.
"""

from spyder.plugins.console.plugin import Console

# The following statement is required to be able to grab internal plugins.
PLUGIN_CLASSES = [Console]
