# -*- coding: utf-8 -*-

# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Kite related widgets."""

from .messagebox import KiteInstallationErrorMessage
from .status import KiteStatusWidget
from .calltoaction import KiteCallToAction
